@@ -8,8 +8,6 @@ use std::sync::{Arc, Mutex};
 
 #[cfg(target_os = "macos")]
 use crossbeam_channel::{unbounded, Sender};
-#[cfg(target_os = "macos")]
-use hvf::MemoryMapping;
 use libc::c_void;
 #[cfg(target_os = "macos")]
 use rutabaga_gfx::RUTABAGA_MEM_HANDLE_TYPE_APPLE;
@@ -28,6 +26,8 @@ use rutabaga_gfx::{
     RUTABAGA_MAP_ACCESS_READ, RUTABAGA_MAP_ACCESS_RW, RUTABAGA_MAP_ACCESS_WRITE,
 };
 use utils::eventfd::EventFd;
+#[cfg(target_os = "macos")]
+use utils::worker_message::WorkerMessage;
 use vm_memory::{GuestAddress, GuestMemory, GuestMemoryMmap, VolatileSlice};
 
 use super::super::Queue as VirtQueue;
@@ -38,7 +38,7 @@ use super::protocol::{
 };
 
 use super::{GpuError, Result};
-use crate::legacy::GicV3;
+use crate::legacy::IrqChip;
 use crate::virtio::fs::ExportTable;
 use crate::virtio::gpu::protocol::VIRTIO_GPU_FLAG_INFO_RING_IDX;
 use crate::virtio::{VirtioShmRegion, VIRTIO_MMIO_INT_VRING};
@@ -107,7 +107,7 @@ pub struct VirtioGpu {
     resources: BTreeMap<u32, VirtioGpuResource>,
     fence_state: Arc<Mutex<FenceState>>,
     #[cfg(target_os = "macos")]
-    map_sender: Sender<MemoryMapping>,
+    map_sender: Sender<WorkerMessage>,
 }
 
 impl VirtioGpu {
@@ -117,7 +117,7 @@ impl VirtioGpu {
         fence_state: Arc<Mutex<FenceState>>,
         interrupt_status: Arc<AtomicUsize>,
         interrupt_evt: EventFd,
-        intc: Option<GicV3>,
+        intc: Option<IrqChip>,
         irq_line: Option<u32>,
     ) -> RutabagaFenceHandler {
         RutabagaFenceHandler::new(move |completed_fence: RutabagaFence| {
@@ -157,9 +157,10 @@ impl VirtioGpu {
 
                     interrupt_status.fetch_or(VIRTIO_MMIO_INT_VRING as usize, Ordering::SeqCst);
                     if let Some(intc) = &intc {
-                        intc.set_irq(irq_line.unwrap());
-                    } else if let Err(e) = interrupt_evt.write(1) {
-                        error!("Failed to signal used queue: {:?}", e);
+                        if let Err(e) = intc.lock().unwrap().set_irq(irq_line, Some(&interrupt_evt))
+                        {
+                            error!("Failed to signal used queue: {:?}", e);
+                        }
                     }
                 } else {
                     i += 1;
@@ -178,10 +179,10 @@ impl VirtioGpu {
         queue_ctl: Arc<Mutex<VirtQueue>>,
         interrupt_status: Arc<AtomicUsize>,
         interrupt_evt: EventFd,
-        intc: Option<GicV3>,
+        intc: Option<IrqChip>,
         irq_line: Option<u32>,
         virgl_flags: u32,
-        #[cfg(target_os = "macos")] map_sender: Sender<MemoryMapping>,
+        #[cfg(target_os = "macos")] map_sender: Sender<WorkerMessage>,
         export_table: Option<ExportTable>,
     ) -> Self {
         let xdg_runtime_dir = match env::var("XDG_RUNTIME_DIR") {
@@ -675,7 +676,7 @@ impl VirtioGpu {
 
                 let (reply_sender, reply_receiver) = unbounded();
                 self.map_sender
-                    .send(MemoryMapping::AddMapping(
+                    .send(WorkerMessage::GpuAddMapping(
                         reply_sender,
                         map_ptr,
                         guest_addr,
@@ -755,7 +756,7 @@ impl VirtioGpu {
 
         let (reply_sender, reply_receiver) = unbounded();
         self.map_sender
-            .send(MemoryMapping::RemoveMapping(
+            .send(WorkerMessage::GpuRemoveMapping(
                 reply_sender,
                 guest_addr,
                 resource.size,
