@@ -4,7 +4,7 @@ use std::path::PathBuf;
 use std::sync::atomic::{AtomicUsize, Ordering};
 use std::sync::{Arc, Mutex, RwLock};
 
-use super::super::super::legacy::GicV3;
+use super::super::super::legacy::IrqChip;
 use super::super::Queue as VirtQueue;
 use super::super::VIRTIO_MMIO_INT_VRING;
 use super::defs;
@@ -107,7 +107,7 @@ pub struct VsockMuxer {
     epoll: Epoll,
     interrupt_evt: EventFd,
     interrupt_status: Arc<AtomicUsize>,
-    intc: Option<GicV3>,
+    intc: Option<IrqChip>,
     irq_line: Option<u32>,
     proxy_map: ProxyMap,
     reaper_sender: Option<Sender<u64>>,
@@ -143,7 +143,7 @@ impl VsockMuxer {
         &mut self,
         mem: GuestMemoryMmap,
         queue: Arc<Mutex<VirtQueue>>,
-        intc: Option<GicV3>,
+        intc: Option<IrqChip>,
         irq_line: Option<u32>,
     ) {
         self.queue = Some(queue.clone());
@@ -242,9 +242,13 @@ impl VsockMuxer {
             self.interrupt_status
                 .fetch_or(VIRTIO_MMIO_INT_VRING as usize, Ordering::SeqCst);
             if let Some(intc) = &self.intc {
-                intc.set_irq(self.irq_line.unwrap());
-            } else if let Err(e) = self.interrupt_evt.write(1) {
-                warn!("failed to signal used queue: {:?}", e);
+                if let Err(e) = intc
+                    .lock()
+                    .unwrap()
+                    .set_irq(self.irq_line, Some(&self.interrupt_evt))
+                {
+                    warn!("failed to signal used queue: {:?}", e);
+                }
             }
         }
     }
@@ -273,7 +277,7 @@ impl VsockMuxer {
             match req._type {
                 defs::SOCK_STREAM => {
                     debug!("vsock: proxy create stream");
-                    let id = (req.peer_port as u64) << 32 | defs::TSI_PROXY_PORT as u64;
+                    let id = ((req.peer_port as u64) << 32) | (defs::TSI_PROXY_PORT as u64);
                     match TcpProxy::new(
                         id,
                         self.cid,
@@ -295,7 +299,7 @@ impl VsockMuxer {
                 }
                 defs::SOCK_DGRAM => {
                     debug!("vsock: proxy create dgram");
-                    let id = (req.peer_port as u64) << 32 | defs::TSI_PROXY_PORT as u64;
+                    let id = ((req.peer_port as u64) << 32) | (defs::TSI_PROXY_PORT as u64);
                     match UdpProxy::new(
                         id,
                         self.cid,
@@ -321,7 +325,7 @@ impl VsockMuxer {
     fn process_connect(&self, pkt: &VsockPacket) {
         debug!("vsock: proxy connect request");
         if let Some(req) = pkt.read_connect_req() {
-            let id = (req.peer_port as u64) << 32 | defs::TSI_PROXY_PORT as u64;
+            let id = ((req.peer_port as u64) << 32) | (defs::TSI_PROXY_PORT as u64);
             debug!("vsock: proxy connect request: id={}", id);
             let update = self
                 .proxy_map
@@ -339,7 +343,7 @@ impl VsockMuxer {
     fn process_getname(&self, pkt: &VsockPacket) {
         debug!("vsock: new getname request");
         if let Some(req) = pkt.read_getname_req() {
-            let id = (req.peer_port as u64) << 32 | (req.local_port as u64);
+            let id = ((req.peer_port as u64) << 32) | (req.local_port as u64);
             debug!(
                 "vsock: new getname request: id={}, peer_port={}, local_port={}",
                 id, req.peer_port, req.local_port
@@ -354,7 +358,7 @@ impl VsockMuxer {
     fn process_sendto_addr(&self, pkt: &VsockPacket) {
         debug!("vsock: new DGRAM sendto addr: src={}", pkt.src_port());
         if let Some(req) = pkt.read_sendto_addr() {
-            let id = (req.peer_port as u64) << 32 | defs::TSI_PROXY_PORT as u64;
+            let id = ((req.peer_port as u64) << 32) | (defs::TSI_PROXY_PORT as u64);
             debug!("vsock: new DGRAM sendto addr: id={}", id);
             let update = self
                 .proxy_map
@@ -370,7 +374,7 @@ impl VsockMuxer {
     }
 
     fn process_sendto_data(&self, pkt: &VsockPacket) {
-        let id = (pkt.src_port() as u64) << 32 | defs::TSI_PROXY_PORT as u64;
+        let id = ((pkt.src_port() as u64) << 32) | (defs::TSI_PROXY_PORT as u64);
         debug!("vsock: DGRAM sendto data: id={} src={}", id, pkt.src_port());
         if let Some(proxy) = self.proxy_map.read().unwrap().get(&id) {
             proxy.lock().unwrap().sendto_data(pkt);
@@ -380,7 +384,7 @@ impl VsockMuxer {
     fn process_listen_request(&self, pkt: &VsockPacket) {
         debug!("vsock: DGRAM listen request: src={}", pkt.src_port());
         if let Some(req) = pkt.read_listen_req() {
-            let id = (req.peer_port as u64) << 32 | defs::TSI_PROXY_PORT as u64;
+            let id = ((req.peer_port as u64) << 32) | (defs::TSI_PROXY_PORT as u64);
             debug!("vsock: DGRAM listen request: id={}", id);
             let update = self
                 .proxy_map
@@ -398,7 +402,7 @@ impl VsockMuxer {
     fn process_accept_request(&self, pkt: &VsockPacket) {
         debug!("vsock: DGRAM accept request: src={}", pkt.src_port());
         if let Some(req) = pkt.read_accept_req() {
-            let id = (req.peer_port as u64) << 32 | defs::TSI_PROXY_PORT as u64;
+            let id = ((req.peer_port as u64) << 32) | (defs::TSI_PROXY_PORT as u64);
             debug!("vsock: DGRAM accept request: id={}", id);
             let update = self
                 .proxy_map
@@ -416,7 +420,7 @@ impl VsockMuxer {
     fn process_proxy_release(&self, pkt: &VsockPacket) {
         debug!("vsock: DGRAM release request: src={}", pkt.src_port());
         if let Some(req) = pkt.read_release_req() {
-            let id = (req.peer_port as u64) << 32 | req.local_port as u64;
+            let id = ((req.peer_port as u64) << 32) | (req.local_port as u64);
             debug!(
                 "vsock: DGRAM release request: id={} local_port={} peer_port={}",
                 id, req.local_port, req.peer_port
@@ -444,7 +448,7 @@ impl VsockMuxer {
 
     fn process_dgram_rw(&self, pkt: &VsockPacket) {
         debug!("vsock: DGRAM OP_RW");
-        let id = (pkt.src_port() as u64) << 32 | defs::TSI_PROXY_PORT as u64;
+        let id = ((pkt.src_port() as u64) << 32) | (defs::TSI_PROXY_PORT as u64);
 
         if let Some(proxy_lock) = self.proxy_map.read().unwrap().get(&id) {
             debug!("vsock: DGRAM allowing OP_RW for {}", pkt.src_port());
@@ -494,11 +498,13 @@ impl VsockMuxer {
 
     fn process_op_request(&mut self, pkt: &VsockPacket) {
         debug!("vsock: OP_REQUEST");
-        let id: u64 = (pkt.src_port() as u64) << 32 | pkt.dst_port() as u64;
+        let id: u64 = ((pkt.src_port() as u64) << 32) | (pkt.dst_port() as u64);
         let mut proxy_map = self.proxy_map.write().unwrap();
 
         if let Some(proxy) = proxy_map.get(&id) {
-            proxy.lock().unwrap().confirm_connect(pkt)
+            if let Some(update) = proxy.lock().unwrap().confirm_connect(pkt) {
+                self.process_proxy_update(id, update);
+            }
         } else if let Some(ref mut ipc_map) = &mut self.unix_ipc_port_map {
             if let Some((path, listen)) = ipc_map.get(&pkt.dst_port()) {
                 let mem = self.mem.as_ref().unwrap();
@@ -540,7 +546,7 @@ impl VsockMuxer {
 
     fn process_op_response(&self, pkt: &VsockPacket) {
         debug!("vsock: OP_RESPONSE");
-        let id: u64 = (pkt.src_port() as u64) << 32 | pkt.dst_port() as u64;
+        let id: u64 = ((pkt.src_port() as u64) << 32) | (pkt.dst_port() as u64);
         let update = self
             .proxy_map
             .read()
@@ -565,7 +571,7 @@ impl VsockMuxer {
 
     fn process_op_shutdown(&self, pkt: &VsockPacket) {
         debug!("vsock: OP_SHUTDOWN");
-        let id: u64 = (pkt.src_port() as u64) << 32 | pkt.dst_port() as u64;
+        let id: u64 = ((pkt.src_port() as u64) << 32) | (pkt.dst_port() as u64);
         if let Some(proxy) = self.proxy_map.read().unwrap().get(&id) {
             proxy.lock().unwrap().shutdown(pkt);
         }
@@ -573,7 +579,7 @@ impl VsockMuxer {
 
     fn process_op_credit_update(&self, pkt: &VsockPacket) {
         debug!("vsock: OP_CREDIT_UPDATE");
-        let id: u64 = (pkt.src_port() as u64) << 32 | pkt.dst_port() as u64;
+        let id: u64 = ((pkt.src_port() as u64) << 32) | (pkt.dst_port() as u64);
         let update = self
             .proxy_map
             .read()
@@ -587,7 +593,7 @@ impl VsockMuxer {
 
     fn process_stream_rw(&self, pkt: &VsockPacket) {
         debug!("vsock: OP_RW");
-        let id: u64 = (pkt.src_port() as u64) << 32 | pkt.dst_port() as u64;
+        let id: u64 = ((pkt.src_port() as u64) << 32) | (pkt.dst_port() as u64);
         if let Some(proxy_lock) = self.proxy_map.read().unwrap().get(&id) {
             debug!(
                 "vsock: allowing OP_RW: src={} dst={}",
@@ -625,7 +631,7 @@ impl VsockMuxer {
 
     fn process_stream_rst(&self, pkt: &VsockPacket) {
         debug!("vsock: OP_RST");
-        let id: u64 = (pkt.src_port() as u64) << 32 | pkt.dst_port() as u64;
+        let id: u64 = ((pkt.src_port() as u64) << 32) | (pkt.dst_port() as u64);
         if let Some(proxy_lock) = self.proxy_map.read().unwrap().get(&id) {
             debug!(
                 "vsock: allowing OP_RST: id={} src={} dst={}",
