@@ -7,7 +7,10 @@
 #[allow(non_snake_case)]
 #[allow(non_upper_case_globals)]
 #[allow(deref_nullptr)]
-mod bindings;
+pub mod bindings;
+
+#[macro_use]
+extern crate log;
 
 use bindings::*;
 
@@ -16,12 +19,11 @@ use std::arch::asm;
 
 use std::convert::TryInto;
 use std::fmt::{Display, Formatter};
-use std::sync::Arc;
+use std::sync::{Arc, LazyLock};
 use std::time::Duration;
 
 #[cfg(all(target_arch = "aarch64", target_os = "macos"))]
-use arch::aarch64::sysreg::{icc_reg_name, SYSREG_MASK};
-use crossbeam_channel::Sender;
+use arch::aarch64::sysreg::{sys_reg_name, SYSREG_MASK};
 use log::debug;
 
 extern "C" {
@@ -37,11 +39,54 @@ const TMR_CTL_IMASK: u64 = 1 << 1;
 const TMR_CTL_ISTATUS: u64 = 1 << 2;
 
 const PSR_MODE_EL1H: u64 = 0x0000_0005;
+const PSR_MODE_EL2H: u64 = 0x0000_0009;
 const PSR_F_BIT: u64 = 0x0000_0040;
 const PSR_I_BIT: u64 = 0x0000_0080;
 const PSR_A_BIT: u64 = 0x0000_0100;
 const PSR_D_BIT: u64 = 0x0000_0200;
-const PSTATE_FAULT_BITS_64: u64 = PSR_MODE_EL1H | PSR_A_BIT | PSR_F_BIT | PSR_I_BIT | PSR_D_BIT;
+const PSTATE_EL1_FAULT_BITS_64: u64 = PSR_MODE_EL1H | PSR_A_BIT | PSR_F_BIT | PSR_I_BIT | PSR_D_BIT;
+const PSTATE_EL2_FAULT_BITS_64: u64 = PSR_MODE_EL2H | PSR_A_BIT | PSR_F_BIT | PSR_I_BIT | PSR_D_BIT;
+
+const HCR_TLOR: u64 = 1 << 35;
+const HCR_RW: u64 = 1 << 31;
+const HCR_TSW: u64 = 1 << 22;
+const HCR_TACR: u64 = 1 << 21;
+const HCR_TIDCP: u64 = 1 << 20;
+const HCR_TSC: u64 = 1 << 19;
+const HCR_TID3: u64 = 1 << 18;
+const HCR_TWE: u64 = 1 << 14;
+const HCR_TWI: u64 = 1 << 13;
+const HCR_BSU_IS: u64 = 1 << 10;
+const HCR_FB: u64 = 1 << 9;
+const HCR_AMO: u64 = 1 << 5;
+const HCR_IMO: u64 = 1 << 4;
+const HCR_FMO: u64 = 1 << 3;
+const HCR_PTW: u64 = 1 << 2;
+const HCR_SWIO: u64 = 1 << 1;
+const HCR_VM: u64 = 1 << 0;
+// Use the same bits as KVM uses in vcpu reset.
+const HCR_EL2_BITS: u64 = HCR_TSC
+    | HCR_TSW
+    | HCR_TWE
+    | HCR_TWI
+    | HCR_VM
+    | HCR_BSU_IS
+    | HCR_FB
+    | HCR_TACR
+    | HCR_AMO
+    | HCR_SWIO
+    | HCR_TIDCP
+    | HCR_RW
+    | HCR_TLOR
+    | HCR_FMO
+    | HCR_IMO
+    | HCR_PTW
+    | HCR_TID3;
+
+const CNTHCTL_EL0VCTEN: u64 = 1 << 1;
+const CNTHCTL_EL0PCTEN: u64 = 1 << 0;
+// Trap accesses to both virtual and physical counter registers.
+const CNTHCTL_EL2_BITS: u64 = CNTHCTL_EL0VCTEN | CNTHCTL_EL0PCTEN;
 
 const EC_WFX_TRAP: u64 = 0x1;
 const EC_AA64_HVC: u64 = 0x16;
@@ -51,10 +96,13 @@ const EC_SYSTEMREGISTERTRAP: u64 = 0x18;
 const EC_DATAABORT: u64 = 0x24;
 const EC_AA64_BKPT: u64 = 0x3c;
 
-#[derive(Clone, Debug)]
+#[derive(Debug)]
 pub enum Error {
+    EnableEL2,
+    FindSymbol(libloading::Error),
     MemoryMap,
     MemoryUnmap,
+    NestedCheck,
     VcpuCreate,
     VcpuInitialRegisters,
     VcpuReadRegister,
@@ -73,8 +121,14 @@ impl Display for Error {
         use self::Error::*;
 
         match self {
+            EnableEL2 => write!(f, "Error enabling EL2 mode in HVF"),
+            FindSymbol(ref err) => write!(f, "Couldn't find symbol in HVF library: {}", err),
             MemoryMap => write!(f, "Error registering memory region in HVF"),
             MemoryUnmap => write!(f, "Error unregistering memory region in HVF"),
+            NestedCheck => write!(
+                f,
+                "Nested virtualization was requested but it's not support in this system"
+            ),
             VcpuCreate => write!(f, "Error creating HVF vCPU instance"),
             VcpuInitialRegisters => write!(f, "Error setting up initial HVF vCPU registers"),
             VcpuReadRegister => write!(f, "Error reading HVF vCPU register"),
@@ -94,12 +148,6 @@ impl Display for Error {
     }
 }
 
-/// Messages for requesting memory maps/unmaps.
-pub enum MemoryMapping {
-    AddMapping(Sender<bool>, u64, u64, u64),
-    RemoveMapping(Sender<bool>, u64, u64),
-}
-
 pub enum InterruptType {
     Irq,
     Fiq,
@@ -154,11 +202,58 @@ pub fn vcpu_set_vtimer_mask(vcpuid: u64, masked: bool) -> Result<(), Error> {
     }
 }
 
+/// Checks if Nested Virtualization is supported on the current system. Only
+/// M3 or newer chips on macOS 15+ will satisfy the requirements.
+pub fn check_nested_virt() -> Result<bool, Error> {
+    type GetEL2Supported =
+        libloading::Symbol<'static, unsafe extern "C" fn(*mut bool) -> hv_return_t>;
+
+    let get_el2_supported: Result<GetEL2Supported, libloading::Error> =
+        unsafe { HVF.get(b"hv_vm_config_get_el2_supported") };
+    if get_el2_supported.is_err() {
+        info!("cannot find hv_vm_config_get_el2_supported symbol");
+        return Ok(false);
+    }
+
+    let mut el2_supported: bool = false;
+    let ret = unsafe { (get_el2_supported.unwrap())(&mut el2_supported) };
+    if ret != HV_SUCCESS {
+        error!("hv_vm_config_get_el2_supported failed: {:?}", ret);
+        return Err(Error::NestedCheck);
+    }
+
+    Ok(el2_supported)
+}
+
 pub struct HvfVm {}
 
+static HVF: LazyLock<libloading::Library> = LazyLock::new(|| unsafe {
+    libloading::Library::new(
+        "/System/Library/Frameworks/Hypervisor.framework/Versions/A/Hypervisor",
+    )
+    .unwrap()
+});
+
 impl HvfVm {
-    pub fn new() -> Result<Self, Error> {
-        let ret = unsafe { hv_vm_create(std::ptr::null_mut()) };
+    pub fn new(nested_enabled: bool) -> Result<Self, Error> {
+        let config = unsafe { hv_vm_config_create() };
+        if nested_enabled {
+            let set_el2_enabled: libloading::Symbol<
+                'static,
+                unsafe extern "C" fn(hv_vm_config_t, bool) -> hv_return_t,
+            > = unsafe {
+                HVF.get(b"hv_vm_config_set_el2_enabled")
+                    .map_err(Error::FindSymbol)?
+            };
+
+            let ret = unsafe { (set_el2_enabled)(config, true) };
+            if ret != HV_SUCCESS {
+                return Err(Error::EnableEL2);
+            }
+        }
+
+        let ret = unsafe { hv_vm_create(config) };
+
         if ret != HV_SUCCESS {
             Err(Error::VmCreate)
         } else {
@@ -176,7 +271,7 @@ impl HvfVm {
             hv_vm_map(
                 host_start_addr as *mut core::ffi::c_void,
                 guest_start_addr,
-                size,
+                size.try_into().unwrap(),
                 (HV_MEMORY_READ | HV_MEMORY_WRITE | HV_MEMORY_EXEC).into(),
             )
         };
@@ -188,7 +283,7 @@ impl HvfVm {
     }
 
     pub fn unmap_memory(&self, guest_start_addr: u64, size: u64) -> Result<(), Error> {
-        let ret = unsafe { hv_vm_unmap(guest_start_addr, size) };
+        let ret = unsafe { hv_vm_unmap(guest_start_addr, size.try_into().unwrap()) };
         if ret != HV_SUCCESS {
             Err(Error::MemoryUnmap)
         } else {
@@ -205,6 +300,7 @@ pub enum VcpuExit<'a> {
     HypervisorCall,
     MmioRead(u64, &'a mut [u8]),
     MmioWrite(u64, &'a [u8]),
+    PsciHandled,
     SecureMonitorCall,
     Shutdown,
     SystemRegister,
@@ -228,10 +324,11 @@ pub struct HvfVcpu<'a> {
     pending_mmio_read: Option<MmioRead>,
     pending_advance_pc: bool,
     vtimer_masked: bool,
+    nested_enabled: bool,
 }
 
 impl HvfVcpu<'_> {
-    pub fn new() -> Result<Self, Error> {
+    pub fn new(mpidr: u64, nested_enabled: bool) -> Result<Self, Error> {
         let mut vcpuid: hv_vcpu_t = 0;
         let vcpu_exit_ptr: *mut hv_vcpu_exit_t = std::ptr::null_mut();
 
@@ -257,8 +354,7 @@ impl HvfVcpu<'_> {
 
         // We write vcpuid to Aff1 as otherwise it won't match the redistributor ID
         // when using HVF in-kernel GICv3.
-        let ret =
-            unsafe { hv_vcpu_set_sys_reg(vcpuid, hv_sys_reg_t_HV_SYS_REG_MPIDR_EL1, vcpuid << 8) };
+        let ret = unsafe { hv_vcpu_set_sys_reg(vcpuid, hv_sys_reg_t_HV_SYS_REG_MPIDR_EL1, mpidr) };
         if ret != HV_SUCCESS {
             return Err(Error::VcpuCreate);
         }
@@ -273,14 +369,43 @@ impl HvfVcpu<'_> {
             pending_mmio_read: None,
             pending_advance_pc: false,
             vtimer_masked: false,
+            nested_enabled,
         })
     }
 
     pub fn set_initial_state(&self, entry_addr: u64, fdt_addr: u64) -> Result<(), Error> {
-        let ret =
-            unsafe { hv_vcpu_set_reg(self.vcpuid, hv_reg_t_HV_REG_CPSR, PSTATE_FAULT_BITS_64) };
-        if ret != HV_SUCCESS {
-            return Err(Error::VcpuInitialRegisters);
+        if self.nested_enabled {
+            let ret = unsafe {
+                hv_vcpu_set_reg(self.vcpuid, hv_reg_t_HV_REG_CPSR, PSTATE_EL2_FAULT_BITS_64)
+            };
+            if ret != HV_SUCCESS {
+                return Err(Error::VcpuInitialRegisters);
+            }
+
+            let ret = unsafe {
+                hv_vcpu_set_sys_reg(self.vcpuid, hv_sys_reg_t_HV_SYS_REG_HCR_EL2, HCR_EL2_BITS)
+            };
+            if ret != HV_SUCCESS {
+                return Err(Error::VcpuInitialRegisters);
+            }
+
+            let ret = unsafe {
+                hv_vcpu_set_sys_reg(
+                    self.vcpuid,
+                    hv_sys_reg_t_HV_SYS_REG_CNTHCTL_EL2,
+                    CNTHCTL_EL2_BITS,
+                )
+            };
+            if ret != HV_SUCCESS {
+                return Err(Error::VcpuInitialRegisters);
+            }
+        } else {
+            let ret = unsafe {
+                hv_vcpu_set_reg(self.vcpuid, hv_reg_t_HV_REG_CPSR, PSTATE_EL1_FAULT_BITS_64)
+            };
+            if ret != HV_SUCCESS {
+                return Err(Error::VcpuInitialRegisters);
+            }
         }
 
         let ret = unsafe { hv_vcpu_set_reg(self.vcpuid, hv_reg_t_HV_REG_PC, entry_addr) };
@@ -346,6 +471,33 @@ impl HvfVcpu<'_> {
         }
     }
 
+    fn handle_psci_request(&self) -> Result<VcpuExit, Error> {
+        match self.read_reg(hv_reg_t_HV_REG_X0)? {
+            0x8400_0000 /* QEMU_PSCI_0_2_FN_PSCI_VERSION */ => {
+                self.write_reg(hv_reg_t_HV_REG_X0, 2)?;
+                Ok(VcpuExit::PsciHandled)
+            },
+            0x8400_0006 /* QEMU_PSCI_0_2_FN_MIGRATE_INFO_TYPE */ => {
+                self.write_reg(hv_reg_t_HV_REG_X0, 2)?;
+                Ok(VcpuExit::PsciHandled)
+            },
+            0x8400_0008 /* QEMU_PSCI_0_2_FN_SYSTEM_OFF */ => {
+                Ok(VcpuExit::Shutdown)
+            },
+            0x8400_0009 /* QEMU_PSCI_0_2_FN_SYSTEM_RESET */ => {
+                Ok(VcpuExit::Shutdown)
+            },
+            0xc400_0003 /* QEMU_PSCI_0_2_FN64_CPU_ON */ => {
+                let mpidr = self.read_reg(hv_reg_t_HV_REG_X1)?;
+                let entry = self.read_reg(hv_reg_t_HV_REG_X2)?;
+                let context_id = self.read_reg(hv_reg_t_HV_REG_X3)?;
+                self.write_reg(hv_reg_t_HV_REG_X0, 0)?;
+                Ok(VcpuExit::CpuOn(mpidr, entry, context_id))
+            }
+            val => panic!("Unexpected val={}", val)
+        }
+    }
+
     pub fn run(&mut self, vcpu_list: Arc<dyn Vcpus>) -> Result<VcpuExit, Error> {
         let pending_irq = vcpu_list.has_pending_irq(self.vcpuid);
 
@@ -456,7 +608,7 @@ impl HvfVcpu<'_> {
                     syndrome,
                     rt,
                     reg,
-                    icc_reg_name(reg).unwrap_or("non-ICC reg")
+                    sys_reg_name(reg).unwrap_or("unknown sysreg")
                 );
 
                 self.pending_advance_pc = true;
@@ -478,7 +630,7 @@ impl HvfVcpu<'_> {
                             "UNKNOWN rt={}, reg={} name={}",
                             rt,
                             reg,
-                            icc_reg_name(reg).unwrap()
+                            sys_reg_name(reg).unwrap_or("unknown sysreg")
                         ),
                     }
                 } else {
@@ -493,8 +645,8 @@ impl HvfVcpu<'_> {
                         panic!(
                             "unexpected write: {} name={}",
                             reg,
-                            icc_reg_name(reg).unwrap_or("non-ICC reg")
-                        )
+                            sys_reg_name(reg).unwrap_or("unknown sysreg")
+                        );
                     }
                 }
             }
@@ -516,35 +668,10 @@ impl HvfVcpu<'_> {
                 let timeout = Duration::from_nanos((cval - now) * (1_000_000_000 / self.cntfrq));
                 Ok(VcpuExit::WaitForEventTimeout(timeout))
             }
-            EC_AA64_HVC => {
-                match self.read_reg(hv_reg_t_HV_REG_X0)? {
-                    0x8400_0000 /* QEMU_PSCI_0_2_FN_PSCI_VERSION */ => {
-                        self.write_reg(hv_reg_t_HV_REG_X0, 2)?;
-                        Ok(VcpuExit::HypervisorCall)
-                    },
-                    0x8400_0006 /* QEMU_PSCI_0_2_FN_MIGRATE_INFO_TYPE */ => {
-                        self.write_reg(hv_reg_t_HV_REG_X0, 2)?;
-                        Ok(VcpuExit::HypervisorCall)
-                    },
-                    0x8400_0008 /* QEMU_PSCI_0_2_FN_SYSTEM_OFF */ => {
-                        Ok(VcpuExit::Shutdown)
-                    },
-                    0x8400_0009 /* QEMU_PSCI_0_2_FN_SYSTEM_RESET */ => {
-                        Ok(VcpuExit::Shutdown)
-                    },
-                    0xc400_0003 /* QEMU_PSCI_0_2_FN64_CPU_ON */ => {
-                        let mpidr = self.read_reg(hv_reg_t_HV_REG_X1)?;
-                        let entry = self.read_reg(hv_reg_t_HV_REG_X2)?;
-                        let context_id = self.read_reg(hv_reg_t_HV_REG_X3)?;
-                        self.write_reg(hv_reg_t_HV_REG_X0, 0)?;
-                        Ok(VcpuExit::CpuOn(mpidr, entry, context_id))
-                    }
-                    val => panic!("Unexpected val={}", val)
-                }
-            }
+            EC_AA64_HVC => self.handle_psci_request(),
             EC_AA64_SMC => {
                 self.pending_advance_pc = true;
-                Ok(VcpuExit::SecureMonitorCall)
+                self.handle_psci_request()
             }
             _ => panic!("unexpected exception: 0x{ec:x}"),
         }
