@@ -8,15 +8,18 @@ use std::env;
 use std::ffi::CStr;
 #[cfg(target_os = "linux")]
 use std::ffi::CString;
+#[cfg(all(target_arch = "x86_64", not(feature = "tee")))]
+use std::fs::File;
 #[cfg(target_os = "linux")]
 use std::os::fd::AsRawFd;
 use std::os::fd::RawFd;
 use std::path::PathBuf;
 use std::slice;
 use std::sync::atomic::{AtomicI32, Ordering};
+#[cfg(not(feature = "efi"))]
+use std::sync::LazyLock;
 use std::sync::Mutex;
 
-#[cfg(target_os = "macos")]
 use crossbeam_channel::unbounded;
 #[cfg(feature = "blk")]
 use devices::virtio::block::ImageType;
@@ -25,8 +28,6 @@ use devices::virtio::net::device::VirtioNetBackend;
 #[cfg(feature = "blk")]
 use devices::virtio::CacheType;
 use env_logger::Env;
-#[cfg(target_os = "macos")]
-use hvf::MemoryMapping;
 #[cfg(not(feature = "efi"))]
 use libc::size_t;
 use libc::{c_char, c_int};
@@ -38,6 +39,8 @@ use vmm::resources::VmResources;
 use vmm::vmm_config::block::BlockDeviceConfig;
 use vmm::vmm_config::boot_source::{BootSourceConfig, DEFAULT_KERNEL_CMDLINE};
 #[cfg(not(feature = "tee"))]
+use vmm::vmm_config::external_kernel::{ExternalKernel, KernelFormat};
+#[cfg(not(feature = "tee"))]
 use vmm::vmm_config::fs::FsDeviceConfig;
 #[cfg(not(feature = "efi"))]
 use vmm::vmm_config::kernel_bundle::KernelBundle;
@@ -48,17 +51,61 @@ use vmm::vmm_config::machine_config::VmConfig;
 use vmm::vmm_config::net::NetworkInterfaceConfig;
 use vmm::vmm_config::vsock::VsockDeviceConfig;
 
-// Minimum krunfw version we require.
-#[cfg(not(feature = "efi"))]
-const KRUNFW_MIN_VERSION: u32 = 4;
 // Value returned on success. We use libc's errors otherwise.
 const KRUN_SUCCESS: i32 = 0;
 // Maximum number of arguments/environment variables we allow
 const MAX_ARGS: usize = 4096;
 
+// krunfw library name for each context
+#[cfg(all(target_os = "linux", not(feature = "amd-sev")))]
+const KRUNFW_NAME: &str = "libkrunfw.so.4";
+#[cfg(all(target_os = "linux", feature = "amd-sev"))]
+const KRUNFW_NAME: &str = "libkrunfw-sev.so.4";
+#[cfg(all(target_os = "macos", not(feature = "efi")))]
+const KRUNFW_NAME: &str = "libkrunfw.4.dylib";
+
 // Path to the init binary to be executed inside the VM.
 const INIT_PATH: &str = "/init.krun";
 
+#[cfg(not(feature = "efi"))]
+static KRUNFW: LazyLock<Option<libloading::Library>> =
+    LazyLock::new(|| unsafe { libloading::Library::new(KRUNFW_NAME).ok() });
+
+#[cfg(not(feature = "efi"))]
+pub struct KrunfwBindings {
+    get_kernel: libloading::Symbol<
+        'static,
+        unsafe extern "C" fn(*mut u64, *mut u64, *mut size_t) -> *mut c_char,
+    >,
+    #[cfg(feature = "tee")]
+    get_initrd: libloading::Symbol<'static, unsafe extern "C" fn(*mut size_t) -> *mut c_char>,
+    #[cfg(feature = "tee")]
+    get_qboot: libloading::Symbol<'static, unsafe extern "C" fn(*mut size_t) -> *mut c_char>,
+}
+
+#[cfg(not(feature = "efi"))]
+impl KrunfwBindings {
+    fn load_bindings() -> Result<KrunfwBindings, libloading::Error> {
+        let krunfw = match KRUNFW.as_ref() {
+            Some(krunfw) => krunfw,
+            None => return Err(libloading::Error::DlOpenUnknown),
+        };
+        Ok(unsafe {
+            KrunfwBindings {
+                get_kernel: krunfw.get(b"krunfw_get_kernel")?,
+                #[cfg(feature = "tee")]
+                get_initrd: krunfw.get(b"krunfw_get_initrd")?,
+                #[cfg(feature = "tee")]
+                get_qboot: krunfw.get(b"krunfw_get_qboot")?,
+            }
+        })
+    }
+
+    pub fn new() -> Option<Self> {
+        Self::load_bindings().ok()
+    }
+}
+
 #[derive(Default)]
 struct TsiConfig {
     port_map: Option<HashMap<u16, u16>>,
@@ -79,6 +126,8 @@ impl Default for NetworkConfig {
 
 #[derive(Default)]
 struct ContextConfig {
+    #[cfg(not(feature = "efi"))]
+    krunfw: Option<KrunfwBindings>,
     vmr: VmResources,
     workdir: Option<String>,
     exec_path: Option<String>,
@@ -101,6 +150,8 @@ struct ContextConfig {
     gpu_shm_size: Option<usize>,
     enable_snd: bool,
     console_output: Option<PathBuf>,
+    vmm_uid: Option<libc::uid_t>,
+    vmm_gid: Option<libc::gid_t>,
 }
 
 impl ContextConfig {
@@ -237,35 +288,19 @@ impl ContextConfig {
     fn set_gpu_shm_size(&mut self, shm_size: usize) {
         self.gpu_shm_size = Some(shm_size);
     }
+
+    fn set_vmm_uid(&mut self, vmm_uid: libc::uid_t) {
+        self.vmm_uid = Some(vmm_uid);
+    }
+
+    fn set_vmm_gid(&mut self, vmm_gid: libc::gid_t) {
+        self.vmm_gid = Some(vmm_gid);
+    }
 }
 
 static CTX_MAP: Lazy<Mutex<HashMap<u32, ContextConfig>>> = Lazy::new(|| Mutex::new(HashMap::new()));
 static CTX_IDS: AtomicI32 = AtomicI32::new(0);
 
-#[cfg(all(not(feature = "tee"), not(feature = "efi")))]
-#[link(name = "krunfw")]
-extern "C" {
-    fn krunfw_get_kernel(
-        load_addr: *mut u64,
-        entry_addr: *mut u64,
-        size: *mut size_t,
-    ) -> *mut c_char;
-    fn krunfw_get_version() -> u32;
-}
-
-#[cfg(feature = "tee")]
-#[link(name = "krunfw-sev")]
-extern "C" {
-    fn krunfw_get_qboot(size: *mut size_t) -> *mut c_char;
-    fn krunfw_get_initrd(size: *mut size_t) -> *mut c_char;
-    fn krunfw_get_kernel(
-        load_addr: *mut u64,
-        entry_addr: *mut u64,
-        size: *mut size_t,
-    ) -> *mut c_char;
-    fn krunfw_get_version() -> u32;
-}
-
 #[no_mangle]
 pub extern "C" fn krun_set_log_level(level: u32) -> i32 {
     let log_level = match level {
@@ -281,70 +316,20 @@ pub extern "C" fn krun_set_log_level(level: u32) -> i32 {
 }
 
 #[no_mangle]
-#[cfg(not(feature = "efi"))]
 pub extern "C" fn krun_create_ctx() -> i32 {
-    let krunfw_version = unsafe { krunfw_get_version() };
-    if krunfw_version < KRUNFW_MIN_VERSION {
-        eprintln!("Unsupported libkrunfw version: {krunfw_version}");
-        return -libc::EINVAL;
-    }
-
-    let mut kernel_guest_addr: u64 = 0;
-    let mut kernel_entry_addr: u64 = 0;
-    let mut kernel_size: usize = 0;
-    let kernel_host_addr = unsafe {
-        krunfw_get_kernel(
-            &mut kernel_guest_addr as *mut u64,
-            &mut kernel_entry_addr as *mut u64,
-            &mut kernel_size as *mut usize,
-        )
-    };
-
-    let mut ctx_cfg = ContextConfig::default();
-
-    let kernel_bundle = KernelBundle {
-        host_addr: kernel_host_addr as u64,
-        guest_addr: kernel_guest_addr,
-        entry_addr: kernel_entry_addr,
-        size: kernel_size,
-    };
-    ctx_cfg.vmr.set_kernel_bundle(kernel_bundle).unwrap();
-
-    #[cfg(feature = "tee")]
-    {
-        let mut qboot_size: usize = 0;
-        let qboot_host_addr = unsafe { krunfw_get_qboot(&mut qboot_size as *mut usize) };
-        let qboot_bundle = QbootBundle {
-            host_addr: qboot_host_addr as u64,
-            size: qboot_size,
+    let ctx_cfg = {
+        let shutdown_efd = if cfg!(feature = "efi") {
+            Some(EventFd::new(utils::eventfd::EFD_NONBLOCK).unwrap())
+        } else {
+            None
         };
-        ctx_cfg.vmr.set_qboot_bundle(qboot_bundle).unwrap();
 
-        let mut initrd_size: usize = 0;
-        let initrd_host_addr = unsafe { krunfw_get_initrd(&mut initrd_size as *mut usize) };
-        let initrd_bundle = InitrdBundle {
-            host_addr: initrd_host_addr as u64,
-            size: initrd_size,
-        };
-        ctx_cfg.vmr.set_initrd_bundle(initrd_bundle).unwrap();
-    }
-
-    let ctx_id = CTX_IDS.fetch_add(1, Ordering::SeqCst);
-    if ctx_id == i32::MAX || CTX_MAP.lock().unwrap().contains_key(&(ctx_id as u32)) {
-        // libkrun is not intended to be used as a daemon for managing VMs.
-        panic!("Context ID namespace exhausted");
-    }
-    CTX_MAP.lock().unwrap().insert(ctx_id as u32, ctx_cfg);
-
-    ctx_id
-}
-
-#[no_mangle]
-#[cfg(feature = "efi")]
-pub extern "C" fn krun_create_ctx() -> i32 {
-    let ctx_cfg = ContextConfig {
-        shutdown_efd: Some(EventFd::new(utils::eventfd::EFD_NONBLOCK).unwrap()),
-        ..Default::default()
+        ContextConfig {
+            #[cfg(not(feature = "efi"))]
+            krunfw: KrunfwBindings::new(),
+            shutdown_efd,
+            ..Default::default()
+        }
     };
 
     let ctx_id = CTX_IDS.fetch_add(1, Ordering::SeqCst);
@@ -521,7 +506,7 @@ pub unsafe extern "C" fn krun_add_disk(
             let cfg = ctx_cfg.get_mut();
             let block_device_config = BlockDeviceConfig {
                 block_id: block_id.to_string(),
-                cache_type: CacheType::Writeback,
+                cache_type: CacheType::auto(disk_path),
                 disk_image_path: disk_path.to_string(),
                 disk_image_format: ImageType::Raw,
                 is_disk_read_only: read_only,
@@ -568,7 +553,7 @@ pub unsafe extern "C" fn krun_add_disk2(
             let cfg = ctx_cfg.get_mut();
             let block_device_config = BlockDeviceConfig {
                 block_id: block_id.to_string(),
-                cache_type: CacheType::Writeback,
+                cache_type: CacheType::auto(disk_path),
                 disk_image_path: disk_path.to_string(),
                 disk_image_format: format,
                 is_disk_read_only: read_only,
@@ -595,7 +580,7 @@ pub unsafe extern "C" fn krun_set_root_disk(ctx_id: u32, c_disk_path: *const c_c
             let cfg = ctx_cfg.get_mut();
             let block_device_config = BlockDeviceConfig {
                 block_id: "root".to_string(),
-                cache_type: CacheType::Writeback,
+                cache_type: CacheType::auto(disk_path),
                 disk_image_path: disk_path.to_string(),
                 disk_image_format: ImageType::Raw,
                 is_disk_read_only: false,
@@ -622,7 +607,7 @@ pub unsafe extern "C" fn krun_set_data_disk(ctx_id: u32, c_disk_path: *const c_c
             let cfg = ctx_cfg.get_mut();
             let block_device_config = BlockDeviceConfig {
                 block_id: "data".to_string(),
-                cache_type: CacheType::Writeback,
+                cache_type: CacheType::auto(disk_path),
                 disk_image_path: disk_path.to_string(),
                 disk_image_format: ImageType::Raw,
                 is_disk_read_only: false,
@@ -1057,6 +1042,52 @@ pub unsafe extern "C" fn krun_set_console_output(ctx_id: u32, c_filepath: *const
     }
 }
 
+#[allow(clippy::missing_safety_doc)]
+#[no_mangle]
+pub unsafe extern "C" fn krun_set_nested_virt(ctx_id: u32, enabled: bool) -> i32 {
+    if enabled && !cfg!(target_os = "macos") {
+        return -libc::EINVAL;
+    }
+
+    match CTX_MAP.lock().unwrap().entry(ctx_id) {
+        Entry::Occupied(mut ctx_cfg) => {
+            let cfg = ctx_cfg.get_mut();
+            cfg.vmr.nested_enabled = enabled;
+            KRUN_SUCCESS
+        }
+        Entry::Vacant(_) => -libc::ENOENT,
+    }
+}
+
+#[allow(clippy::missing_safety_doc)]
+#[no_mangle]
+pub unsafe extern "C" fn krun_check_nested_virt() -> i32 {
+    #[cfg(target_os = "macos")]
+    match hvf::check_nested_virt() {
+        Ok(supp) => supp as i32,
+        Err(_) => -libc::EINVAL,
+    }
+
+    #[cfg(not(target_os = "macos"))]
+    -libc::EOPNOTSUPP
+}
+
+#[allow(clippy::missing_safety_doc)]
+#[no_mangle]
+pub extern "C" fn krun_split_irqchip(ctx_id: u32, enable: bool) -> i32 {
+    if enable && !cfg!(target_arch = "x86_64") {
+        return -libc::EINVAL;
+    }
+    match CTX_MAP.lock().unwrap().entry(ctx_id) {
+        Entry::Occupied(mut ctx_cfg) => {
+            let cfg = ctx_cfg.get_mut();
+            cfg.vmr.split_irqchip = enable;
+            KRUN_SUCCESS
+        }
+        Entry::Vacant(_) => -libc::ENOENT,
+    }
+}
+
 #[allow(clippy::missing_safety_doc)]
 #[no_mangle]
 pub unsafe extern "C" fn krun_set_smbios_oem_strings(
@@ -1104,6 +1135,217 @@ fn create_virtio_net(ctx_cfg: &mut ContextConfig, backend: VirtioNetBackend) {
         .expect("Failed to create network interface");
 }
 
+#[cfg(all(target_arch = "x86_64", not(feature = "tee")))]
+fn map_kernel(ctx_id: u32, kernel_path: &PathBuf) -> i32 {
+    let file = match File::options().read(true).write(false).open(kernel_path) {
+        Ok(file) => file,
+        Err(err) => {
+            error!("Error opening external kernel: {err}");
+            return -libc::EINVAL;
+        }
+    };
+
+    let kernel_size = file.metadata().unwrap().len();
+
+    let kernel_host_addr = unsafe {
+        libc::mmap(
+            std::ptr::null_mut(),
+            kernel_size as usize,
+            libc::PROT_READ,
+            libc::MAP_SHARED,
+            file.as_raw_fd(),
+            0_i64,
+        )
+    };
+    if std::ptr::eq(kernel_host_addr, libc::MAP_FAILED) {
+        error!("Can't load kernel into process map");
+        return -libc::EINVAL;
+    }
+
+    let kernel_bundle = KernelBundle {
+        host_addr: kernel_host_addr as u64,
+        guest_addr: 0x8000_0000,
+        entry_addr: 0x8000_0000,
+        size: kernel_size as usize,
+    };
+
+    match CTX_MAP.lock().unwrap().entry(ctx_id) {
+        Entry::Occupied(mut ctx_cfg) => ctx_cfg
+            .get_mut()
+            .vmr
+            .set_kernel_bundle(kernel_bundle)
+            .unwrap(),
+        Entry::Vacant(_) => return -libc::ENOENT,
+    }
+
+    KRUN_SUCCESS
+}
+
+#[cfg(feature = "tee")]
+#[allow(clippy::format_collect)]
+#[allow(clippy::missing_safety_doc)]
+#[no_mangle]
+pub unsafe extern "C" fn krun_set_kernel(_ctx_id: u32, _c_kernel_path: *const c_char) -> i32 {
+    -libc::EOPNOTSUPP
+}
+
+#[cfg(not(feature = "tee"))]
+#[allow(clippy::format_collect)]
+#[allow(clippy::missing_safety_doc)]
+#[no_mangle]
+pub unsafe extern "C" fn krun_set_kernel(
+    ctx_id: u32,
+    c_kernel_path: *const c_char,
+    kernel_format: u32,
+    c_initramfs_path: *const c_char,
+    c_cmdline: *const c_char,
+) -> i32 {
+    let path = match CStr::from_ptr(c_kernel_path).to_str() {
+        Ok(path) => PathBuf::from(path),
+        Err(e) => {
+            error!("Error parsing kernel_path: {:?}", e);
+            return -libc::EINVAL;
+        }
+    };
+
+    let format = match kernel_format {
+        // For raw kernels in x86_64, we map the kernel into the
+        // process and treat it as a bundled kernel.
+        #[cfg(all(target_arch = "x86_64", not(feature = "tee")))]
+        0 => return map_kernel(ctx_id, &path),
+        #[cfg(target_arch = "aarch64")]
+        0 => KernelFormat::Raw,
+        1 => KernelFormat::Elf,
+        2 => KernelFormat::PeGz,
+        3 => KernelFormat::ImageBz2,
+        4 => KernelFormat::ImageGz,
+        5 => KernelFormat::ImageZstd,
+        _ => {
+            return -libc::EINVAL;
+        }
+    };
+
+    let (initramfs_path, initramfs_size) = if !c_initramfs_path.is_null() {
+        match CStr::from_ptr(c_initramfs_path).to_str() {
+            Ok(path) => {
+                let path = PathBuf::from(path);
+                let size = match std::fs::metadata(&path) {
+                    Ok(metadata) => metadata.len(),
+                    Err(e) => {
+                        error!("Can't read initramfs metadata: {:?}", e);
+                        return -libc::EINVAL;
+                    }
+                };
+                (Some(path), size)
+            }
+            Err(e) => {
+                error!("Error parsing initramfs path: {:?}", e);
+                return -libc::EINVAL;
+            }
+        }
+    } else {
+        (None, 0)
+    };
+
+    let cmdline = if !c_cmdline.is_null() {
+        match CStr::from_ptr(c_cmdline).to_str() {
+            Ok(cmdline) => Some(cmdline.to_string()),
+            Err(e) => {
+                error!("Error parsing kernel cmdline: {:?}", e);
+                return -libc::EINVAL;
+            }
+        }
+    } else {
+        None
+    };
+
+    let external_kernel = ExternalKernel {
+        path,
+        format,
+        initramfs_path,
+        initramfs_size,
+        cmdline,
+    };
+
+    match CTX_MAP.lock().unwrap().entry(ctx_id) {
+        Entry::Occupied(mut ctx_cfg) => ctx_cfg.get_mut().vmr.set_external_kernel(external_kernel),
+        Entry::Vacant(_) => return -libc::ENOENT,
+    }
+
+    KRUN_SUCCESS
+}
+
+#[cfg(not(feature = "efi"))]
+unsafe fn load_krunfw_payload(
+    krunfw: &KrunfwBindings,
+    vmr: &mut VmResources,
+) -> Result<(), libloading::Error> {
+    let mut kernel_guest_addr: u64 = 0;
+    let mut kernel_entry_addr: u64 = 0;
+    let mut kernel_size: usize = 0;
+    let kernel_host_addr = unsafe {
+        (krunfw.get_kernel)(
+            &mut kernel_guest_addr as *mut u64,
+            &mut kernel_entry_addr as *mut u64,
+            &mut kernel_size as *mut usize,
+        )
+    };
+    let kernel_bundle = KernelBundle {
+        host_addr: kernel_host_addr as u64,
+        guest_addr: kernel_guest_addr,
+        entry_addr: kernel_entry_addr,
+        size: kernel_size,
+    };
+    vmr.set_kernel_bundle(kernel_bundle).unwrap();
+
+    #[cfg(feature = "tee")]
+    {
+        let mut qboot_size: usize = 0;
+        let qboot_host_addr = unsafe { (krunfw.get_qboot)(&mut qboot_size as *mut usize) };
+        let qboot_bundle = QbootBundle {
+            host_addr: qboot_host_addr as u64,
+            size: qboot_size,
+        };
+        vmr.set_qboot_bundle(qboot_bundle).unwrap();
+
+        let mut initrd_size: usize = 0;
+        let initrd_host_addr = unsafe { (krunfw.get_initrd)(&mut initrd_size as *mut usize) };
+        let initrd_bundle = InitrdBundle {
+            host_addr: initrd_host_addr as u64,
+            size: initrd_size,
+        };
+        vmr.set_initrd_bundle(initrd_bundle).unwrap();
+    }
+
+    Ok(())
+}
+
+#[no_mangle]
+pub extern "C" fn krun_setuid(ctx_id: u32, uid: libc::uid_t) -> i32 {
+    match CTX_MAP.lock().unwrap().entry(ctx_id) {
+        Entry::Occupied(mut ctx_cfg) => {
+            let cfg = ctx_cfg.get_mut();
+            cfg.set_vmm_uid(uid);
+        }
+        Entry::Vacant(_) => return -libc::ENOENT,
+    }
+
+    KRUN_SUCCESS
+}
+
+#[no_mangle]
+pub extern "C" fn krun_setgid(ctx_id: u32, gid: libc::gid_t) -> i32 {
+    match CTX_MAP.lock().unwrap().entry(ctx_id) {
+        Entry::Occupied(mut ctx_cfg) => {
+            let cfg = ctx_cfg.get_mut();
+            cfg.set_vmm_gid(gid);
+        }
+        Entry::Vacant(_) => return -libc::ENOENT,
+    }
+
+    KRUN_SUCCESS
+}
+
 #[no_mangle]
 pub extern "C" fn krun_start_enter(ctx_id: u32) -> i32 {
     #[cfg(target_os = "linux")]
@@ -1128,6 +1370,19 @@ pub extern "C" fn krun_start_enter(ctx_id: u32) -> i32 {
         None => return -libc::ENOENT,
     };
 
+    #[cfg(not(feature = "efi"))]
+    if ctx_cfg.vmr.external_kernel.is_none() && ctx_cfg.vmr.kernel_bundle.is_none() {
+        if let Some(ref krunfw) = ctx_cfg.krunfw {
+            if let Err(err) = unsafe { load_krunfw_payload(krunfw, &mut ctx_cfg.vmr) } {
+                eprintln!("Can't load libkrunfw symbols: {err}");
+                return -libc::ENOENT;
+            }
+        } else {
+            eprintln!("Couldn't find or load {}", KRUNFW_NAME);
+            return -libc::ENOENT;
+        }
+    }
+
     #[cfg(feature = "blk")]
     for block_cfg in ctx_cfg.get_block_cfg() {
         if ctx_cfg.vmr.add_block_device(block_cfg).is_err() {
@@ -1222,14 +1477,26 @@ pub extern "C" fn krun_start_enter(ctx_id: u32) -> i32 {
         ctx_cfg.vmr.set_console_output(console_output);
     }
 
-    #[cfg(target_os = "macos")]
-    let (sender, receiver) = unbounded();
+    if let Some(gid) = ctx_cfg.vmm_gid {
+        if unsafe { libc::setgid(gid) } != 0 {
+            error!("Failed to set gid {}", gid);
+            return -std::io::Error::last_os_error().raw_os_error().unwrap();
+        }
+    }
+
+    if let Some(uid) = ctx_cfg.vmm_uid {
+        if unsafe { libc::setuid(uid) } != 0 {
+            error!("Failed to set uid {}", uid);
+            return -std::io::Error::last_os_error().raw_os_error().unwrap();
+        }
+    }
+
+    let (sender, _receiver) = unbounded();
 
     let _vmm = match vmm::builder::build_microvm(
         &ctx_cfg.vmr,
         &mut event_manager,
         ctx_cfg.shutdown_efd,
-        #[cfg(target_os = "macos")]
         sender,
     ) {
         Ok(vmm) => vmm,
@@ -1240,25 +1507,17 @@ pub extern "C" fn krun_start_enter(ctx_id: u32) -> i32 {
     };
 
     #[cfg(target_os = "macos")]
-    let mapper_vmm = _vmm.clone();
+    if ctx_cfg.gpu_virgl_flags.is_some() {
+        vmm::worker::start_worker_thread(_vmm.clone(), _receiver).unwrap();
+    }
 
-    #[cfg(target_os = "macos")]
-    std::thread::Builder::new()
-        .name("mapping worker".into())
-        .spawn(move || loop {
-            match receiver.recv() {
-                Err(e) => error!("Error in receiver: {:?}", e),
-                Ok(m) => match m {
-                    MemoryMapping::AddMapping(s, h, g, l) => {
-                        mapper_vmm.lock().unwrap().add_mapping(s, h, g, l)
-                    }
-                    MemoryMapping::RemoveMapping(s, g, l) => {
-                        mapper_vmm.lock().unwrap().remove_mapping(s, g, l)
-                    }
-                },
-            }
-        })
-        .unwrap();
+    #[cfg(target_arch = "x86_64")]
+    if ctx_cfg.vmr.split_irqchip {
+        vmm::worker::start_worker_thread(_vmm.clone(), _receiver.clone()).unwrap();
+    }
+
+    #[cfg(feature = "amd-sev")]
+    vmm::worker::start_worker_thread(_vmm.clone(), _receiver.clone()).unwrap();
 
     loop {
         match event_manager.run() {
