@@ -10,8 +10,8 @@ use libc::{c_int, c_void, siginfo_t};
 use std::cell::Cell;
 use std::fmt::{Display, Formatter};
 use std::io;
+use std::ops::Range;
 
-#[cfg(feature = "tee")]
 use std::os::unix::io::RawFd;
 
 use std::result;
@@ -22,9 +22,6 @@ use std::thread;
 
 use super::super::{FC_EXIT_CODE_GENERIC_ERROR, FC_EXIT_CODE_OK};
 
-#[cfg(feature = "amd-sev")]
-use super::tee::amdsev::{AmdSev, Error as SevError};
-
 #[cfg(feature = "amd-sev")]
 use super::tee::amdsnp::{AmdSnp, Error as SnpError};
 
@@ -34,29 +31,33 @@ use kbs_types::Tee;
 #[cfg(feature = "tee")]
 use crate::resources::TeeConfig;
 use crate::vmm_config::machine_config::CpuFeaturesTemplate;
-#[cfg(target_arch = "aarch64")]
-use arch::aarch64::gic::GICDevice;
 #[cfg(target_arch = "x86_64")]
 use cpuid::{c3, filter_cpuid, t2, VmSpec};
 #[cfg(target_arch = "x86_64")]
 use kvm_bindings::{
-    kvm_clock_data, kvm_debugregs, kvm_irqchip, kvm_lapic_state, kvm_mp_state, kvm_pit_config,
-    kvm_pit_state2, kvm_regs, kvm_sregs, kvm_vcpu_events, kvm_xcrs, kvm_xsave, CpuId, MsrList,
-    Msrs, KVM_CLOCK_TSC_STABLE, KVM_IRQCHIP_IOAPIC, KVM_IRQCHIP_PIC_MASTER, KVM_IRQCHIP_PIC_SLAVE,
-    KVM_MAX_CPUID_ENTRIES, KVM_PIT_SPEAKER_DUMMY,
+    kvm_clock_data, kvm_debugregs, kvm_irqchip, kvm_lapic_state, kvm_mp_state, kvm_pit_state2,
+    kvm_regs, kvm_sregs, kvm_vcpu_events, kvm_xcrs, kvm_xsave, CpuId, MsrList, Msrs,
+    KVM_CLOCK_TSC_STABLE, KVM_IRQCHIP_IOAPIC, KVM_IRQCHIP_PIC_MASTER, KVM_IRQCHIP_PIC_SLAVE,
+    KVM_MAX_CPUID_ENTRIES,
 };
-use kvm_bindings::{kvm_userspace_memory_region, KVM_API_VERSION};
-use kvm_ioctls::*;
+use kvm_bindings::{
+    kvm_create_guest_memfd, kvm_memory_attributes, kvm_userspace_memory_region,
+    kvm_userspace_memory_region2, KVM_API_VERSION, KVM_MEMORY_ATTRIBUTE_PRIVATE,
+    KVM_MEM_GUEST_MEMFD, KVM_SYSTEM_EVENT_RESET, KVM_SYSTEM_EVENT_SHUTDOWN,
+};
+#[cfg(feature = "tee")]
+use kvm_bindings::{kvm_enable_cap, KVM_CAP_EXIT_HYPERCALL, KVM_MEMORY_EXIT_FLAG_PRIVATE};
+use kvm_ioctls::{Cap::*, *};
 use utils::eventfd::EventFd;
 use utils::signal::{register_signal_handler, sigrtmin, Killable};
 use utils::sm::StateMachine;
+#[cfg(feature = "tee")]
+use utils::worker_message::{MemoryProperties, WorkerMessage};
 use vm_memory::{
     Address, GuestAddress, GuestMemory, GuestMemoryError, GuestMemoryMmap, GuestMemoryRegion,
+    GuestRegionMmap,
 };
 
-#[cfg(feature = "amd-sev")]
-use sev::launch::sev as sev_launch;
-
 #[cfg(feature = "amd-sev")]
 use sev::launch::snp;
 
@@ -69,6 +70,8 @@ pub enum Error {
     #[cfg(target_arch = "x86_64")]
     /// A call to cpuid instruction failed.
     CpuId(cpuid::Error),
+    /// Unable to create a KVM guest_memfd.
+    CreateGuestMemfd(kvm_ioctls::Error),
     #[cfg(target_arch = "x86_64")]
     /// Error configuring the floating point related registers
     FPUConfiguration(arch::x86_64::regs::Error),
@@ -79,6 +82,9 @@ pub enum Error {
     GuestMSRs(arch::x86_64::msr::Error),
     /// Hyperthreading flag is not initialized.
     HTNotInitialized,
+    /// Unable to enable KVM hypercall exits.
+    #[cfg(feature = "tee")]
+    HypercallExitEnable(kvm_ioctls::Error),
     /// Cannot configure the IRQ.
     Irq(kvm_ioctls::Error),
     /// The host kernel reports an invalid KVM API version.
@@ -105,23 +111,13 @@ pub enum Error {
     #[cfg(target_arch = "x86_64")]
     /// Error configuring the general purpose registers
     REGSConfiguration(arch::x86_64::regs::Error),
-    #[cfg(target_arch = "aarch64")]
-    /// Error setting up the global interrupt controller.
-    SetupGIC(arch::aarch64::gic::Error),
+    /// Cannot set memory region attributes.
+    SetMemoryAttributes(kvm_ioctls::Error),
     /// Cannot set the memory regions.
     SetUserMemoryRegion(kvm_ioctls::Error),
     /// Error creating memory map for SHM region.
     ShmMmap(io::Error),
     #[cfg(feature = "amd-sev")]
-    /// Error initializing the Secure Virtualization Backend (SEV).
-    SevSecVirtInit(SevError),
-    #[cfg(feature = "amd-sev")]
-    /// Error preparing the VM for Secure Virtualization (SEV).
-    SevSecVirtPrepare(SevError),
-    #[cfg(feature = "amd-sev")]
-    /// Error attesting the Secure VM (SEV).
-    SevSecVirtAttest(SevError),
-    #[cfg(feature = "amd-sev")]
     /// Error initializing the Secure Virtualization Backend (SNP).
     SnpSecVirtInit(SnpError),
     #[cfg(feature = "amd-sev")]
@@ -215,6 +211,9 @@ pub enum Error {
     VcpuTlsNotPresent,
     /// Unexpected KVM_RUN exit reason
     VcpuUnhandledKvmExit,
+    /// Unsupported KVM_EXIT_HYPERCALL.
+    #[cfg(feature = "tee")]
+    VcpuUnsupportedHypercall,
     /// Cannot open the VM file descriptor.
     VmFd(kvm_ioctls::Error),
     #[cfg(target_arch = "x86_64")]
@@ -246,10 +245,13 @@ impl Display for Error {
         match self {
             #[cfg(target_arch = "x86_64")]
             CpuId(e) => write!(f, "Cpuid error: {e:?}"),
+            CreateGuestMemfd(e) => write!(f, "Unable to create KVM guest_memfd: {e:?}"),
             GuestMemoryMmap(e) => write!(f, "Guest memory error: {e:?}"),
             #[cfg(target_arch = "x86_64")]
             GuestMSRs(e) => write!(f, "Retrieving supported guest MSRs fails: {e:?}"),
             HTNotInitialized => write!(f, "Hyperthreading flag is not initialized"),
+            #[cfg(feature = "tee")]
+            HypercallExitEnable(e) => write!(f, "Unable to enable KVM hypercall exits: {e}"),
             KvmApiVersion(v) => {
                 write!(f, "The host kernel reports an invalid KVM API version: {v}")
             }
@@ -270,23 +272,9 @@ impl Display for Error {
                 f,
                 "Cannot set the local interruption due to bad configuration: {e:?}"
             ),
+            SetMemoryAttributes(e) => write!(f, "Cannot set memory region attributes: {e}"),
             SetUserMemoryRegion(e) => write!(f, "Cannot set the memory regions: {e}"),
             ShmMmap(e) => write!(f, "Error creating memory map for SHM region: {e}"),
-            #[cfg(feature = "tee")]
-            SevSecVirtInit(e) => {
-                write!(
-                    f,
-                    "Error initializing the Secure Virtualization Backend (SEV): {e:?}"
-                )
-            }
-            #[cfg(feature = "tee")]
-            SevSecVirtPrepare(e) => write!(
-                f,
-                "Error preparing the VM for Secure Virtualization (SEV): {e:?}"
-            ),
-            #[cfg(feature = "tee")]
-            SevSecVirtAttest(e) => write!(f, "Error attesting the Secure VM (SEV): {e:?}"),
-
             #[cfg(feature = "tee")]
             SnpSecVirtInit(e) => write!(
                 f,
@@ -366,6 +354,8 @@ impl Display for Error {
             VcpuTlsInit => write!(f, "Cannot clean init vcpu TLS"),
             VcpuTlsNotPresent => write!(f, "Vcpu not present in TLS"),
             VcpuUnhandledKvmExit => write!(f, "Unexpected KVM_RUN exit reason"),
+            #[cfg(feature = "tee")]
+            VcpuUnsupportedHypercall => write!(f, "Unsupported KVM_EXIT_HYPERCALL"),
             #[cfg(target_arch = "x86_64")]
             VmGetPit2(e) => write!(f, "Failed to get KVM vm pit state: {e}"),
             #[cfg(target_arch = "x86_64")]
@@ -379,8 +369,6 @@ impl Display for Error {
             #[cfg(target_arch = "x86_64")]
             VmSetIrqChip(e) => write!(f, "Failed to set KVM vm irqchip: {e}"),
             #[cfg(target_arch = "aarch64")]
-            SetupGIC(e) => write!(f, "Error setting up the global interrupt controller: {e:?}"),
-            #[cfg(target_arch = "aarch64")]
             VcpuArmPreferredTarget(e) => {
                 write!(f, "Error getting the Vcpu preferred target on Arm: {e}")
             }
@@ -413,7 +401,6 @@ pub struct KvmContext {
 
 impl KvmContext {
     pub fn new() -> Result<Self> {
-        use kvm_ioctls::Cap::*;
         let kvm = Kvm::new().expect("Error creating the Kvm object");
 
         // Check that KVM has the correct version.
@@ -463,19 +450,13 @@ pub struct Vm {
     #[cfg(target_arch = "x86_64")]
     supported_msrs: MsrList,
 
-    // Arm specific fields.
-    // On aarch64 we need to keep around the fd obtained by creating the VGIC device.
-    #[cfg(target_arch = "aarch64")]
-    irqchip_handle: Option<Box<dyn GICDevice>>,
+    #[cfg(feature = "amd-sev")]
+    tee: Option<AmdSnp>,
 
     #[cfg(feature = "amd-sev")]
-    sev: Option<AmdSev>,
+    pub tee_config: Tee,
 
-    #[cfg(feature = "amd-sev")]
-    snp: Option<AmdSnp>,
-
-    #[cfg(feature = "amd-sev")]
-    pub tee: Tee,
+    pub guest_memfds: Vec<(Range<u64>, RawFd)>,
 }
 
 impl Vm {
@@ -500,15 +481,16 @@ impl Vm {
             supported_cpuid,
             #[cfg(target_arch = "x86_64")]
             supported_msrs,
-            #[cfg(target_arch = "aarch64")]
-            irqchip_handle: None,
+            guest_memfds: Vec::new(),
         })
     }
 
     #[cfg(feature = "amd-sev")]
     pub fn new(kvm: &Kvm, tee_config: &TeeConfig) -> Result<Self> {
         //create fd for interacting with kvm-vm specific functions
-        let vm_fd = kvm.create_vm().map_err(Error::VmFd)?;
+        let vm_fd = kvm
+            .create_vm_with_type(4 /* KVM_X86_SNP_VM */)
+            .map_err(Error::VmFd)?;
 
         let supported_cpuid = kvm
             .get_supported_cpuid(KVM_MAX_CPUID_ENTRIES)
@@ -517,12 +499,17 @@ impl Vm {
         let supported_msrs =
             arch::x86_64::msr::supported_guest_msrs(kvm).map_err(Error::GuestMSRs)?;
 
-        let (sev, snp) = match tee_config.tee {
-            Tee::Sev => (
-                Some(AmdSev::new(tee_config).map_err(Error::SevSecVirtInit)?),
-                None,
-            ),
-            Tee::Snp => (None, Some(AmdSnp::new().map_err(Error::SnpSecVirtInit)?)),
+        let cap = kvm_enable_cap {
+            cap: KVM_CAP_EXIT_HYPERCALL,
+            flags: 0,
+            args: [1 << 12 /* KVM_HC_MAP_GPA_RANGE */, 0, 0, 0],
+            ..Default::default()
+        };
+
+        vm_fd.enable_cap(&cap).map_err(Error::HypercallExitEnable)?;
+
+        let tee = match tee_config.tee {
+            Tee::Snp => Some(AmdSnp::new().map_err(Error::SnpSecVirtInit)?),
             _ => return Err(Error::InvalidTee),
         };
 
@@ -531,9 +518,9 @@ impl Vm {
             next_mem_slot: 0,
             supported_cpuid,
             supported_msrs,
-            sev,
-            snp,
-            tee: tee_config.tee,
+            tee,
+            tee_config: tee_config.tee,
+            guest_memfds: Vec::new(),
         })
     }
 
@@ -558,25 +545,9 @@ impl Vm {
         if guest_mem.num_regions() > kvm_max_memslots {
             return Err(Error::NotEnoughMemorySlots);
         }
+
         for region in guest_mem.iter() {
-            // It's safe to unwrap because the guest address is valid.
-            let host_addr = guest_mem.get_host_address(region.start_addr()).unwrap();
-            debug!("Guest memory starts at {:x?}", host_addr);
-            let memory_region = kvm_userspace_memory_region {
-                slot: self.next_mem_slot,
-                guest_phys_addr: region.start_addr().raw_value(),
-                memory_size: region.len(),
-                userspace_addr: host_addr as u64,
-                flags: 0,
-            };
-            // Safe because we mapped the memory region, we made sure that the regions
-            // are not overlapping.
-            unsafe {
-                self.fd
-                    .set_user_memory_region(memory_region)
-                    .map_err(Error::SetUserMemoryRegion)?;
-            };
-            self.next_mem_slot += 1;
+            self.memory_region_set(guest_mem, region)?;
         }
 
         #[cfg(target_arch = "x86_64")]
@@ -587,32 +558,89 @@ impl Vm {
         Ok(())
     }
 
-    #[cfg(feature = "amd-sev")]
-    pub fn sev_secure_virt_prepare(
-        &mut self,
-        guest_mem: &GuestMemoryMmap,
-    ) -> Result<sev_launch::Launcher<sev_launch::Started, RawFd, RawFd>> {
-        match &self.sev {
-            Some(s) => s
-                .vm_prepare(&self.fd, guest_mem)
-                .map_err(Error::SevSecVirtPrepare),
-            None => Err(Error::InvalidTee),
+    pub fn guest_memfd_get(&self, gpa: u64) -> Option<(RawFd, u64)> {
+        for (range, rawfd) in self.guest_memfds.iter() {
+            if range.contains(&gpa) {
+                return Some((*rawfd, range.start));
+            }
         }
+        None
     }
 
-    #[cfg(feature = "amd-sev")]
-    pub fn sev_secure_virt_attest(
-        &self,
+    #[allow(unused_mut)]
+    fn memory_region_set(
+        &mut self,
         guest_mem: &GuestMemoryMmap,
-        measured_regions: Vec<MeasuredRegion>,
-        launcher: sev_launch::Launcher<sev_launch::Started, RawFd, RawFd>,
+        region: &GuestRegionMmap,
     ) -> Result<()> {
-        match &self.sev {
-            Some(s) => s
-                .vm_attest(&self.fd, guest_mem, measured_regions, launcher)
-                .map_err(Error::SevSecVirtAttest),
-            None => Err(Error::InvalidTee),
+        let host_addr = guest_mem.get_host_address(region.start_addr()).unwrap();
+        let start = region.start_addr().raw_value();
+        let end = start + region.len();
+
+        if !self.fd.check_extension(GuestMemfd) {
+            let memory_region = kvm_userspace_memory_region {
+                slot: self.next_mem_slot,
+                guest_phys_addr: start,
+                memory_size: region.len(),
+                userspace_addr: host_addr as u64,
+                flags: 0,
+            };
+
+            // Safe because we mapped the memory region, we made sure that the regions
+            // are not overlapping.
+            unsafe {
+                self.fd
+                    .set_user_memory_region(memory_region)
+                    .map_err(Error::SetUserMemoryRegion)?;
+            };
+        } else {
+            // Create a guest_memfd and set the region.
+            let guest_memfd = self
+                .fd
+                .create_guest_memfd(kvm_create_guest_memfd {
+                    size: region.size() as u64,
+                    flags: 0,
+                    reserved: [0; 6],
+                })
+                .map_err(Error::CreateGuestMemfd)?;
+
+            let memory_region = kvm_userspace_memory_region2 {
+                slot: self.next_mem_slot,
+                flags: KVM_MEM_GUEST_MEMFD,
+                guest_phys_addr: start,
+                memory_size: region.len(),
+                userspace_addr: host_addr as u64,
+                guest_memfd_offset: 0,
+                guest_memfd: guest_memfd as u32,
+                pad1: 0,
+                pad2: [0; 14],
+            };
+
+            // Safe because we mapped the memory region, we made sure that the regions
+            // are not overlapping.
+            unsafe {
+                self.fd
+                    .set_user_memory_region2(memory_region)
+                    .map_err(Error::SetUserMemoryRegion)?;
+            };
+
+            let attr = kvm_memory_attributes {
+                address: start,
+                size: region.len(),
+                attributes: KVM_MEMORY_ATTRIBUTE_PRIVATE as u64,
+                flags: 0,
+            };
+
+            self.fd
+                .set_memory_attributes(attr)
+                .map_err(Error::SetMemoryAttributes)?;
+
+            self.guest_memfds.push((Range { start, end }, guest_memfd));
         }
+
+        self.next_mem_slot += 1;
+
+        Ok(())
     }
 
     #[cfg(feature = "amd-sev")]
@@ -620,7 +648,7 @@ impl Vm {
         &self,
         guest_mem: &GuestMemoryMmap,
     ) -> Result<snp::Launcher<snp::Started, RawFd, RawFd>> {
-        match &self.snp {
+        match &self.tee {
             Some(s) => s
                 .vm_prepare(&self.fd, guest_mem)
                 .map_err(Error::SnpSecVirtPrepare),
@@ -629,14 +657,14 @@ impl Vm {
     }
 
     #[cfg(feature = "amd-sev")]
-    pub fn snp_secure_virt_attest(
+    pub fn snp_secure_virt_measure(
         &self,
         cpuid: CpuId,
         guest_mem: &GuestMemoryMmap,
         measured_regions: Vec<MeasuredRegion>,
         launcher: snp::Launcher<snp::Started, RawFd, RawFd>,
     ) -> Result<()> {
-        match &self.snp {
+        match &self.tee {
             Some(s) => s
                 .vm_measure(cpuid, guest_mem, measured_regions, launcher)
                 .map_err(Error::SnpSecVirtAttest),
@@ -644,35 +672,6 @@ impl Vm {
         }
     }
 
-    /// Creates the irq chip and an in-kernel device model for the PIT.
-    #[cfg(target_arch = "x86_64")]
-    pub fn setup_irqchip(&self) -> Result<()> {
-        self.fd.create_irq_chip().map_err(Error::VmSetup)?;
-        let pit_config = kvm_pit_config {
-            // We need to enable the emulation of a dummy speaker port stub so that writing to port
-            // 0x61 (i.e. KVM_SPEAKER_BASE_ADDRESS) does not trigger an exit to user space.
-            flags: KVM_PIT_SPEAKER_DUMMY,
-            ..Default::default()
-        };
-        self.fd.create_pit2(pit_config).map_err(Error::VmSetup)
-    }
-
-    /// Creates the GIC (Global Interrupt Controller).
-    #[cfg(target_arch = "aarch64")]
-    pub fn setup_irqchip(&mut self, vcpu_count: u8) -> Result<()> {
-        self.irqchip_handle = Some(
-            arch::aarch64::gic::create_gic(&self.fd, vcpu_count.into()).map_err(Error::SetupGIC)?,
-        );
-        Ok(())
-    }
-
-    /// Gets a reference to the irqchip of the VM
-    #[allow(clippy::borrowed_box)]
-    #[cfg(target_arch = "aarch64")]
-    pub fn get_irqchip(&self) -> &Box<dyn GICDevice> {
-        self.irqchip_handle.as_ref().unwrap()
-    }
-
     /// Gets a reference to the kvm file descriptor owned by this VM.
     pub fn fd(&self) -> &VmFd {
         &self.fd
@@ -794,6 +793,9 @@ pub struct Vcpu {
     response_receiver: Option<Receiver<VcpuResponse>>,
     // The transmitting end of the responses channel owned by the vcpu side.
     response_sender: Sender<VcpuResponse>,
+
+    #[cfg(feature = "tee")]
+    pm_sender: Sender<WorkerMessage>,
 }
 
 impl Vcpu {
@@ -825,7 +827,7 @@ impl Vcpu {
         // _before_ running this, then there is nothing we can do.
         Self::TLS_VCPU_PTR.with(|cell: &VcpuCell| {
             if let Some(vcpu_ptr) = cell.get() {
-                if vcpu_ptr == self as *mut Vcpu {
+                if std::ptr::eq(vcpu_ptr, self) {
                     Self::TLS_VCPU_PTR.with(|cell: &VcpuCell| cell.take());
                     return Ok(());
                 }
@@ -897,6 +899,7 @@ impl Vcpu {
         msr_list: MsrList,
         io_bus: devices::Bus,
         exit_evt: EventFd,
+        #[cfg(feature = "tee")] pm_sender: Sender<WorkerMessage>,
     ) -> Result<Self> {
         let kvm_vcpu = vm_fd.create_vcpu(id as u64).map_err(Error::VcpuFd)?;
         let (event_sender, event_receiver) = unbounded();
@@ -915,6 +918,8 @@ impl Vcpu {
             event_sender: Some(event_sender),
             response_receiver: Some(response_receiver),
             response_sender,
+            #[cfg(feature = "tee")]
+            pm_sender,
         })
     }
 
@@ -1175,9 +1180,11 @@ impl Vcpu {
         self.fd
             .set_sregs(&state.sregs)
             .map_err(Error::VcpuSetSregs)?;
-        self.fd
-            .set_xsave(&state.xsave)
-            .map_err(Error::VcpuSetXsave)?;
+        unsafe {
+            self.fd
+                .set_xsave(&state.xsave)
+                .map_err(Error::VcpuSetXsave)?;
+        }
         self.fd.set_xcrs(&state.xcrs).map_err(Error::VcpuSetXcrs)?;
         self.fd
             .set_debug_regs(&state.debug_regs)
@@ -1198,6 +1205,35 @@ impl Vcpu {
     fn run_emulation(&mut self) -> Result<VcpuEmulation> {
         match self.fd.run() {
             Ok(run) => match run {
+                #[cfg(feature = "tee")]
+                VcpuExit::Hypercall(hypercall) => {
+                    if hypercall.nr != 12
+                    /* KVM_HC_MAP_GPA_RANGE */
+                    {
+                        return Err(Error::VcpuUnsupportedHypercall);
+                    }
+
+                    let gpa = hypercall.args[0];
+                    let size = hypercall.args[1] * 0x1000; /* TARGET_PAGE_SIZE */
+                    let attributes = hypercall.args[2];
+
+                    let private = !matches!(attributes, 0);
+
+                    let mem_properties = MemoryProperties { gpa, size, private };
+
+                    let (response_sender, response_receiver) = unbounded();
+                    self.pm_sender
+                        .send(WorkerMessage::ConvertMemory(
+                            response_sender.clone(),
+                            mem_properties,
+                        ))
+                        .unwrap();
+                    if !response_receiver.recv().unwrap() {
+                        error!("Unable to convert memory with properties: gpa: 0x{:x} size: 0x{:x} to_private: {}", gpa, size, private);
+                        return Err(Error::VcpuUnhandledKvmExit);
+                    }
+                    Ok(VcpuEmulation::Handled)
+                }
                 #[cfg(target_arch = "x86_64")]
                 VcpuExit::IoIn(addr, data) => {
                     self.io_bus.read(0, u64::from(addr), data);
@@ -1208,6 +1244,26 @@ impl Vcpu {
                     self.io_bus.write(0, u64::from(addr), data);
                     Ok(VcpuEmulation::Handled)
                 }
+                #[cfg(feature = "tee")]
+                VcpuExit::MemoryFault { gpa, size, flags } => {
+                    let private = (flags & (KVM_MEMORY_EXIT_FLAG_PRIVATE as u64)) != 0;
+
+                    let mem_properties = MemoryProperties { gpa, size, private };
+
+                    let (response_sender, response_receiver) = unbounded();
+                    self.pm_sender
+                        .send(WorkerMessage::ConvertMemory(
+                            response_sender.clone(),
+                            mem_properties,
+                        ))
+                        .unwrap();
+                    if !response_receiver.recv().unwrap() {
+                        error!("Unable to convert memory with properties: gpa: 0x{:x} size: 0x{:x} to_private: {}", gpa, size, private);
+                        return Err(Error::VcpuUnhandledKvmExit);
+                    }
+
+                    Ok(VcpuEmulation::Handled)
+                }
                 VcpuExit::MmioRead(addr, data) => {
                     if let Some(ref mmio_bus) = self.mmio_bus {
                         mmio_bus.read(0, addr, data);
@@ -1238,6 +1294,14 @@ impl Vcpu {
                     error!("Received KVM_EXIT_INTERNAL_ERROR signal");
                     Err(Error::VcpuUnhandledKvmExit)
                 }
+                VcpuExit::SystemEvent(event, _reason) => {
+                    match event {
+                        KVM_SYSTEM_EVENT_SHUTDOWN => info!("Received KVM_SYSTEM_EVENT_SHUTDOWN"),
+                        KVM_SYSTEM_EVENT_RESET => info!("Received KVM_SYSTEM_EVENT_RESET"),
+                        _ => error!("Received an unexpected System Event: {event}"),
+                    }
+                    Ok(VcpuEmulation::Stopped)
+                }
                 r => {
                     // TODO: Are we sure we want to finish running a vcpu upon
                     // receiving a vm exit that is not necessarily an error?
@@ -1489,6 +1553,7 @@ mod tests {
 
     use super::*;
     use devices;
+    use devices::legacy::KvmIoapic;
 
     use utils::signal::validate_signal_num;
 
@@ -1510,6 +1575,7 @@ mod tests {
         let kvm = KvmContext::new().unwrap();
         let gm = GuestMemoryMmap::from_ranges(&[(GuestAddress(0), mem_size)]).unwrap();
         let mut vm = Vm::new(kvm.fd()).expect("Cannot create new vm");
+        let _kvmioapic = KvmIoapic::new(&vm.fd()).unwrap();
         assert!(vm.memory_init(&gm, kvm.max_memslots()).is_ok());
 
         let exit_evt = EventFd::new(utils::eventfd::EFD_NONBLOCK).unwrap();
@@ -1517,7 +1583,6 @@ mod tests {
         let vcpu;
         #[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
         {
-            vm.setup_irqchip().unwrap();
             vcpu = Vcpu::new_x86_64(
                 1,
                 vm.fd(),
@@ -1531,7 +1596,6 @@ mod tests {
         #[cfg(target_arch = "aarch64")]
         {
             vcpu = Vcpu::new_aarch64(1, vm.fd(), exit_evt).unwrap();
-            vm.setup_irqchip(1).expect("Cannot setup irqchip");
         }
 
         (vm, vcpu, gm)
@@ -1578,50 +1642,6 @@ mod tests {
         assert!(vm.memory_init(&gm, kvm_context.max_memslots()).is_err());
     }
 
-    #[cfg(target_arch = "x86_64")]
-    #[test]
-    fn test_setup_irqchip() {
-        let kvm_context = KvmContext::new().unwrap();
-        let vm = Vm::new(kvm_context.fd()).expect("Cannot create new vm");
-
-        vm.setup_irqchip().expect("Cannot setup irqchip");
-        // Trying to setup two irqchips will result in EEXIST error. At the moment
-        // there is no good way of testing the actual error because io::Error does not implement
-        // PartialEq.
-        assert!(vm.setup_irqchip().is_err());
-
-        let _vcpu = Vcpu::new_x86_64(
-            1,
-            vm.fd(),
-            vm.supported_cpuid().clone(),
-            vm.supported_msrs().clone(),
-            devices::Bus::new(),
-            EventFd::new(utils::eventfd::EFD_NONBLOCK).unwrap(),
-        )
-        .unwrap();
-        // Trying to setup irqchip after KVM_VCPU_CREATE was called will result in error.
-        assert!(vm.setup_irqchip().is_err());
-    }
-
-    #[cfg(target_arch = "aarch64")]
-    #[test]
-    fn test_setup_irqchip() {
-        let kvm = KvmContext::new().unwrap();
-
-        let mut vm = Vm::new(kvm.fd()).expect("Cannot create new vm");
-        let vcpu_count = 1;
-        let _vcpu = Vcpu::new_aarch64(
-            1,
-            vm.fd(),
-            EventFd::new(utils::eventfd::EFD_NONBLOCK).unwrap(),
-        )
-        .unwrap();
-
-        vm.setup_irqchip(vcpu_count).expect("Cannot setup irqchip");
-        // Trying to setup two irqchips will result in EEXIST error.
-        assert!(vm.setup_irqchip(vcpu_count).is_err());
-    }
-
     #[cfg(target_arch = "x86_64")]
     #[test]
     fn test_configure_vcpu() {
